@@ -1,12 +1,14 @@
 #pragma once
 
-#include "launch.hpp"
+#include "../taskflow.hpp"
 
 namespace tf {
 
 // Function: make_for_each_task
 template <typename B, typename E, typename C, typename P = DefaultPartitioner>
 auto make_for_each_task(B b, E e, C c, P part = P()) {
+  
+  using namespace std::string_literals;
 
   using B_t = std::decay_t<unwrap_ref_decay_t<B>>;
   using E_t = std::decay_t<unwrap_ref_decay_t<E>>;
@@ -20,14 +22,15 @@ auto make_for_each_task(B b, E e, C c, P part = P()) {
     size_t W = rt.executor().num_workers();
     size_t N = std::distance(beg, end);
 
-    // only myself - no need to spawn another graph
+    // the workload is sequentially doable
     if(W <= 1 || N <= part.chunk_size()) {
-      launch_loop(part, [&](){
-        std::for_each(beg, end, c);
-      });
+      part([=](){ std::for_each(beg, end, c); })();
       return;
     }
-
+    
+    PreemptionGuard preemption_guard(rt);
+    
+    // use no more workers than the iteration count
     if(N < W) {
       W = N;
     }
@@ -35,11 +38,11 @@ auto make_for_each_task(B b, E e, C c, P part = P()) {
     // static partitioner
     if constexpr(part.type() == PartitionerType::STATIC) {
       size_t chunk_size;
-      for(size_t w=0, curr_b=0; w<W && curr_b < N; ++w, curr_b += chunk_size) {
+      for(size_t w=0, curr_b=0; w<W && curr_b < N;) {
         chunk_size = part.adjusted_chunk_size(N, W, w);
-        launch_loop(W, w, rt, part, [=, &c, &part] () mutable {
+        auto task = part([=] () mutable {
           part.loop(N, W, curr_b, chunk_size,
-            [&, prev_e=size_t{0}](size_t part_b, size_t part_e) mutable {
+            [=, prev_e=size_t{0}](size_t part_b, size_t part_e) mutable {
               std::advance(beg, part_b - prev_e);
               for(size_t x = part_b; x<part_e; x++) {
                 c(*beg++);
@@ -48,24 +51,26 @@ auto make_for_each_task(B b, E e, C c, P part = P()) {
             }
           ); 
         });
+        (++w == W || (curr_b += chunk_size) >= N) ? task() : rt.silent_async(task);
       }
-
-      rt.corun_all();
     }
     // dynamic partitioner
     else {
-      std::atomic<size_t> next(0);
-      launch_loop(N, W, rt, next, part, [=, &c, &next, &part] () mutable {
-        part.loop(N, W, next, 
-          [&, prev_e=size_t{0}](size_t part_b, size_t part_e) mutable {
-            std::advance(beg, part_b - prev_e);
-            for(size_t x = part_b; x<part_e; x++) {
-              c(*beg++);
+      auto next = std::make_shared<std::atomic<size_t>>(0);
+      for(size_t w=0; w<W;) {
+        auto task = part([=] () mutable {
+          part.loop(N, W, *next, 
+            [=, prev_e=size_t{0}](size_t part_b, size_t part_e) mutable {
+              std::advance(beg, part_b - prev_e);
+              for(size_t x = part_b; x<part_e; x++) {
+                c(*beg++);
+              }
+              prev_e = part_e;
             }
-            prev_e = part_e;
-          }
-        );
-      });
+          );
+        });
+        (++w == W) ? task() : rt.silent_async(task);
+      }
     }
 
   };
@@ -74,6 +79,8 @@ auto make_for_each_task(B b, E e, C c, P part = P()) {
 // Function: make_for_each_index_task
 template <typename B, typename E, typename S, typename C, typename P = DefaultPartitioner>
 auto make_for_each_index_task(B b, E e, S s, C c, P part = P()){
+  
+  using namespace std::string_literals;
 
   using B_t = std::decay_t<unwrap_ref_decay_t<B>>;
   using E_t = std::decay_t<unwrap_ref_decay_t<E>>;
@@ -96,14 +103,16 @@ auto make_for_each_index_task(B b, E e, S s, C c, P part = P()){
 
     // only myself - no need to spawn another graph
     if(W <= 1 || N <= part.chunk_size()) {
-      launch_loop(part, [&](){
+      part([&](){
         for(size_t x=0; x<N; x++, beg+=inc) {
           c(beg);
         }
-      });
+      })();
       return;
     }
 
+    PreemptionGuard preemption_guard(rt);
+    
     if(N < W) {
       W = N;
     }
@@ -111,38 +120,34 @@ auto make_for_each_index_task(B b, E e, S s, C c, P part = P()){
     // static partitioner
     if constexpr(part.type() == PartitionerType::STATIC) {
       size_t chunk_size;
-      for(size_t w=0, curr_b=0; w<W && curr_b < N; ++w, curr_b += chunk_size) {
+      for(size_t w=0, curr_b=0; w<W && curr_b < N;) {
         chunk_size = part.adjusted_chunk_size(N, W, w);
-        launch_loop(W, w, rt, part, [=, &c, &part] () mutable {
-          part.loop(N, W, curr_b, chunk_size,
-            [&](size_t part_b, size_t part_e) {
-              auto idx = static_cast<B_t>(part_b) * inc + beg;
-              for(size_t x=part_b; x<part_e; x++, idx += inc) {
-                c(idx);
-              }
-            }
-          );
-        });
-      }
-
-      rt.corun_all();
-    }
-    // dynamic partitioner
-    else {
-      std::atomic<size_t> next(0);
-      launch_loop(N, W, rt, next, part, [=, &c, &next, &part] () mutable {
-        part.loop(N, W, next, 
-          [&](size_t part_b, size_t part_e) {
+        auto task = part([=] () mutable {
+          part.loop(N, W, curr_b, chunk_size, [=] (size_t part_b, size_t part_e) {
             auto idx = static_cast<B_t>(part_b) * inc + beg;
             for(size_t x=part_b; x<part_e; x++, idx += inc) {
               c(idx);
             }
-          }
-        );
-      });
+          });
+        });
+        (++w == W || (curr_b += chunk_size) >= N) ? task() : rt.silent_async(task);
+      }
+    }
+    // dynamic partitioner
+    else {
+      auto next = std::make_shared<std::atomic<size_t>>(0);
+      for(size_t w=0; w<W;) {
+        auto task = part([=] () mutable {
+          part.loop(N, W, *next, [=] (size_t part_b, size_t part_e) {
+            auto idx = static_cast<B_t>(part_b) * inc + beg;
+            for(size_t x=part_b; x<part_e; x++, idx += inc) {
+              c(idx);
+            }
+          });
+        });
+        (++w == W) ? task() : rt.silent_async(task);
+      }
     }
-
-
   };
 }
 
@@ -156,9 +161,6 @@ Task FlowBuilder::for_each(B beg, E end, C c, P part) {
   return emplace(
     make_for_each_task(beg, end, c, part)
   );
-  //return Task(_graph._emplace_back("", 0, nullptr, nullptr, 0,
-  //  std::in_place_type_t<Node::Static>{}, std::forward<C>(c)
-  //));
 }
 
 // ----------------------------------------------------------------------------
